from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import reduce
from typing import Optional, List, Callable, Tuple, TypeVar, Generic, Union
from loquax.abstractions.constants import Constants

# This can be any type that the Rule is supposed to operate on, e.g., Syllable or Phoneme
T = TypeVar("T")


class Tokenizer(ABC):
    """
    Abstract base class for a tokenizer.
    """

    @abstractmethod
    def tokenize(self, text: str) -> List[str]:
        pass


@dataclass
class Rule(Generic[T]):
    """
    Represents a single rule that can be used to check if a given unit (e.g. syllable, phoneme)
    matches certain criteria. The rule can be defined using a check function, a specific value,
    or by being a wildcard.
    """

    check_fn: Optional[Callable[[T], bool]] = None
    val: Optional[str] = None
    wildcard: bool = False

    def matches(self, unit: T) -> bool:
        if self.wildcard:
            return True

        match_predicate = (
            lambda s: s.__repr__() == self.val
            if self.val is not None
            else self.check_fn(s)
        )
        return match_predicate(unit)


@dataclass
class RuleSequence(Generic[T]):
    """
    Represents a sequence of rules that can be used to match against a sequence of units (e.g. syllables, phonemes).
    The sequence is matched when all individual rules in the sequence match their corresponding units.
    """

    rules: List[Rule[T]]

    def matches(self, units: List[T]) -> bool:
        if len(units) < len(self.rules):
            return False

        return all(
            rule.matches(unit)
            for rule, unit in zip(self.rules, units[-len(self.rules) :])
        )


@dataclass
class Morphism(Generic[T]):
    """
    Represents a transformation applied to units in a sequence based on
    specific criteria, such as the surrounding units or the presence of a certain pattern.
    """

    target: Rule[T]
    transformation: Union[Callable[[T], T], T]
    prefix: Optional[RuleSequence[T]] = None
    suffix: Optional[RuleSequence[T]] = None

    def matches(self, units: List[T], index: int) -> bool:
        return (
            (
                self.prefix.matches(
                    units[max(0, index - len(self.prefix.rules)) : index]
                )
                if self.prefix
                else True
            )
            and self.target.matches(units[index])
            and (
                self.suffix.matches(
                    units[index + 1 : index + 1 + len(self.suffix.rules)]
                )
                if self.suffix
                else True
            )
        )

    def apply(self, units: List[T]) -> List[T]:
        return [
            (
                self.transformation(unit)
                if callable(self.transformation)
                else self.transformation
            )
            if self.matches(units, i)
            else unit
            for i, unit in enumerate(units)
        ]


@dataclass
class MorphismStore(Generic[T]):
    """
    Represents a store for morphisms that can be applied to a sequence of units.
    The store holds a list of morphisms and provides a method to apply all morphisms to a given sequence.
    """

    morphisms: List[Morphism[T]]

    def apply_all(self, units: List[T]) -> List[T]:
        return reduce(lambda units, morph: morph.apply(units), self.morphisms, units)


@dataclass
class PhonemeSyllabificationRuleStore:
    """
    Represents a store for phoneme rules that are used in syllabification.
    These rules help to determine how to split a sequence of phonemes into syllables.
    """

    rules: List[RuleSequence["Phoneme"]]

    def apply_all(
        self, phonemes: List["Phoneme"]
    ) -> Tuple[List["Phoneme"], List["Phoneme"]]:
        if not phonemes:
            return [], []

        matching_rule = next(
            (
                rule
                for rule in self.rules
                if len(rule.rules) <= len(phonemes)
                and rule.matches(phonemes[-len(rule.rules) :])
            ),
            None,
        )

        if matching_rule is None:
            return [], []

        rule_length = len(matching_rule.rules)
        return phonemes[:-rule_length], phonemes[-rule_length:]


@dataclass
class Language:
    """
    Represents a language, containing information such as language name, constants, syllabification rules,
    morphisms, and an optional ISO 639 code. The Language class provides an organized way to store
    language-specific data for processing phoneme sequences.
    """

    language_name: str
    iso_639_code: str
    constants: Constants
    syllabification_rules: PhonemeSyllabificationRuleStore
    syllable_morphisms: MorphismStore["Syllable"]
    phoneme_morphisms: MorphismStore["Phoneme"]
    tokenizer: Tokenizer


@dataclass
class Phoneme:
    """
    Class to represent a single phoneme or sound unit in a word.
    """

    val: str
    lang: Language
    ipa: Optional[str] = None

    def __post_init__(self):
        if self.val not in self.lang.constants.equivalencies:
            raise ValueError(
                f"The symbol '{self.val}' is not a valid phoneme in the language: '{self.lang.language_name}'."
            )
        # If ipa is not set, use the first IPA equivalent from equivalencies
        if self.ipa is None:
            self.ipa = self.lang.constants.equivalencies[self.val][0]
        elif self.ipa not in self.lang.constants.equivalencies[self.val]:
            raise ValueError(
                f"""The IPA symbol '{self.ipa}' is not a valid equivalent for phoneme '{self.val}' 
                in the language: '{self.lang.language_name}'."""
            )

    @property
    def is_diphthong(self) -> bool:
        return self.val in self.lang.constants.diphthongs

    @property
    def is_stop(self) -> bool:
        return self.val in self.lang.constants.stop_letters

    @property
    def is_liquid(self) -> bool:
        return self.val in self.lang.constants.liquid_letters

    @property
    def is_aspirate(self) -> bool:
        return self.val in self.lang.constants.aspirates

    @property
    def is_vowel(self) -> bool:
        return self.val in self.lang.constants.vowels

    @property
    def is_consonant(self) -> bool:
        return self.val in self.lang.constants.consonants

    def set_ipa(self, new_ipa: str) -> "Phoneme":
        # Ensure the new IPA is a valid equivalent for this phoneme
        if new_ipa not in self.lang.constants.equivalencies[self.val]:
            raise ValueError(
                f"""The IPA symbol '{new_ipa}' is not a valid equivalent for phoneme '{self.val}' 
                in the language: '{self.lang.language_name}'."""
            )
        # Return a new Phoneme with the same val and lang, but with the new IPA
        return Phoneme(val=self.val, lang=self.lang, ipa=new_ipa)

    def to_str(self, ipa: bool = False):
        if ipa:
            return self.val
        else:
            return self.ipa

    def __repr__(self) -> str:
        return self.val

    def __str__(self) -> str:
        return self.__repr__()


@dataclass
class Syllable:
    """
    Class to represent a syllable in a word.
    Syllable: List of Phonemes with an Onset, Nucleus, and Coda.
    O-N-C (Nucleus is a vowel, Onset and Coda are optional groups of consonants).
    """

    phonemes: List[Phoneme]
    lang: Language
    is_long: bool = False

    def __post_init__(self):
        num_vowels = len([p for p in self.phonemes if p.is_vowel])
        if num_vowels > 1:
            raise ValueError(
                f"Invalid syllable: {self.phonemes} contains more than one vowel phoneme."
            )

    @property
    def nucleus(self) -> Optional[List[Phoneme]]:
        first_vowel_index = self._find_first_vowel_index()
        if first_vowel_index is not None:
            return [self.phonemes[first_vowel_index]]
        return None

    @property
    def onset(self) -> Optional[List[Phoneme]]:
        first_vowel_index = self._find_first_vowel_index()
        if first_vowel_index is not None:
            return self.phonemes[:first_vowel_index]
        return self.phonemes

    @property
    def coda(self) -> Optional[List[Phoneme]]:
        first_vowel_index = self._find_first_vowel_index()
        if first_vowel_index is not None:
            return self.phonemes[first_vowel_index + 1 :]
        return None

    @property
    def val(self) -> str:
        return "".join([phoneme.val for phoneme in self.phonemes])

    def _find_first_vowel_index(self) -> Optional[int]:
        return next((i for i, p in enumerate(self.phonemes) if p.is_vowel), None)

    def scansion_str(self, ipa: bool = False) -> str:
        symbol = "U" if not self.is_long else "—"
        return symbol.center(len(self.__repr__() if ipa else self.__str__()))

    def to_str(self, ipa: bool = False) -> str:
        return "".join(
            [phoneme.ipa if ipa else phoneme.val for phoneme in self.phonemes]
        )

    def __repr__(self):
        return "".join([phoneme.val for phoneme in self.phonemes])

    def __str__(self):
        return self.__repr__()



from loquax.abstractions.constants import Constants
from loquax.languages.latin_conf.constants import *
from loquax.languages.latin_conf.rules import (
    latin_syllabification_rules,
    latin_syllable_morphisms,
    latin_phoneme_morphisms,
)
from loquax.abstractions import Language
from loquax.languages.latin_conf.tokenizer import LatinTokenizer

Latin = Language(
    language_name="Classical Latin",
    iso_639_code="lat",
    constants=Constants(
        vowel_equivalencies=latin_vowel_equivalencies,
        consonant_equivalencies=latin_consonant_equivalencies,
        aspirates=latin_aspirates,
        diphthongs=latin_diphthongs,
        liquid_letters=latin_liquid_letters,
        stop_letters=latin_stop_letters,
    ),
    syllabification_rules=latin_syllabification_rules,
    syllable_morphisms=latin_syllable_morphisms,
    phoneme_morphisms=latin_phoneme_morphisms,
    tokenizer=LatinTokenizer(),
)



from typing import List

from loquax.abstractions import Language, Syllable, Phoneme
from loquax.abstractions.phonology import get_phonemes


def get_syllables_from_phonemes(token: List[Phoneme], lang: Language) -> List[Syllable]:
    # Helper function to create a Syllable based on current, previous, and next vowel indices
    def _create_syllable(
        tkn: List[Phoneme],
        current_index: int,
        prev_vowel_index: int,
        next_vowel_index: int,
    ):
        # Define the consonant ranges between the current and adjacent vowels (if they exist)
        prev_consonant_range = tkn[prev_vowel_index + 1 : vowel_indices[current_index]]
        next_consonant_range = tkn[vowel_indices[current_index] + 1 : next_vowel_index]

        # Apply syllabification rules to the consonant ranges
        rules = lang.syllabification_rules
        onset = rules.apply_all(prev_consonant_range)[1]
        coda = rules.apply_all(next_consonant_range)[0]

        # Create a Syllable based on the current index and syllabification rules
        match current_index:
            case 0:
                return Syllable(tkn[: vowel_indices[current_index] + 1] + coda, lang)
            case index if index == len(vowel_indices) - 1:
                return Syllable(onset + tkn[vowel_indices[current_index] :], lang)
            case _:
                return Syllable(
                    onset + [tkn[vowel_indices[current_index]]] + coda, lang
                )

    # Get the indices of all vowels in the token
    vowel_indices = [i for i, phoneme in enumerate(token) if phoneme.is_vowel]

    # Create syllables based on the number of vowels and their positions
    match len(vowel_indices):
        case num_vowels if num_vowels <= 1:
            return lang.syllable_morphisms.apply_all([Syllable(token, lang)])
        case _:
            syllables: List[Syllable] = [
                _create_syllable(
                    token,
                    index,
                    vowel_indices[index - 1] if index > 0 else 0,
                    vowel_indices[index + 1]
                    if index < len(vowel_indices) - 1
                    else len(token),
                )
                for index in range(len(vowel_indices))
            ]

            return lang.syllable_morphisms.apply_all(syllables)


def get_syllables_from_token(token: str, lang: Language) -> List[Syllable]:
    return lang.syllable_morphisms.apply_all(
        get_syllables_from_phonemes(get_phonemes(token, lang), lang)
    )



from dataclasses import dataclass
from itertools import chain
from typing import Dict, List, Set


@dataclass
class Constants:
    vowel_equivalencies: Dict[str, List[str]]
    consonant_equivalencies: Dict[str, List[str]]
    aspirates: Set[str]
    diphthongs: Set[str]
    liquid_letters: Set[str]
    stop_letters: Set[str]

    def get_distinct_elements(self, d: Dict) -> set:
        """
        Get the set of distinct elements from the keys and values of a dictionary.
        If a value is a list, its elements will be included in the output set.

        :param d (dict): the input dictionary
        :return: the set of distinct elements
        """
        return set(
            chain(d.keys(), *(v if isinstance(v, list) else [v] for v in d.values()))
        )

    @property
    def equivalencies(self) -> Dict[str, List[str]]:
        return self.vowel_equivalencies | self.consonant_equivalencies

    @property
    def vowels(self) -> Set[str]:
        return set(self.get_distinct_elements(self.vowel_equivalencies))

    @property
    def consonants(self) -> Set[str]:
        return set(self.get_distinct_elements(self.consonant_equivalencies))

    @property
    def symbol_groups(self) -> Set[str]:
        return set(
            filter(
                lambda x: len(x) > 1,
                set(
                    self.aspirates.union(
                        self.diphthongs,
                        self.liquid_letters,
                        self.stop_letters,
                        self.vowels,
                        self.consonants,
                    )
                ),
            )
        )



from dataclasses import replace

from loquax.abstractions import (
    PhonemeSyllabificationRuleStore,
    MorphismStore,
    Rule,
    Syllable,
    RuleSequence,
    Morphism,
    Phoneme,
)
from loquax.text_processing import has_macron


"""
SYLLABIFICATION RULES

(http://www.wheelockslatin.com/chapters/introduction/introduction_syllables.html) 
"""

latin_syllabification_rules = PhonemeSyllabificationRuleStore(
    [
        # SYLLABIFICATION RULE 1: A stop (p, b, t, d, c, g) plus a liquid (l,r) generally count as a single consonant
        # and go with the following vowel
        RuleSequence[Phoneme](
            [
                Rule[Phoneme](lambda p: p.is_stop),
                Rule[Phoneme](lambda p: p.is_liquid),
            ]
        ),
        # SYLLABIFICATION RULE 2: Counted as single consonants are qu and the aspirates ch, ph, th, which should
        # never be separated in syllabification
        RuleSequence[Phoneme]([Rule[Phoneme](lambda p: p.is_aspirate)]),
        # SYLLABIFICATION RULE 3:  A single consonant between two vowels goes with the second vowel
        RuleSequence[Phoneme]([Rule[Phoneme](lambda p: p.is_consonant)]),
    ]
)

"""
SYLLABLE LENGTH RULES

(http://www.wheelockslatin.com/chapters/introduction/introduction_syllables.html) 
"""
long_nature_morphism = Morphism[Syllable](
    target=Rule[Syllable](
        check_fn=lambda s: s.nucleus
        and any(has_macron(p) or p.is_diphthong for p in s.nucleus)
    ),
    transformation=lambda s: replace(s, is_long=True),
)

long_position_morphism_1 = Morphism[Syllable](
    target=Rule[Syllable](check_fn=lambda s: s.nucleus and s.coda and len(s.coda) >= 2),
    transformation=lambda s: replace(s, is_long=True),
)

long_position_morphism_2 = Morphism[Syllable](
    target=Rule[Syllable](check_fn=lambda s: s.nucleus and s.coda and len(s.coda) >= 1),
    transformation=lambda s: replace(s, is_long=True),
    suffix=RuleSequence(
        [Rule[Syllable](check_fn=lambda s: s.coda and len(s.onset) >= 1)]
    ),
)

long_position_morphism_3 = Morphism[Syllable](
    target=Rule[Syllable](check_fn=lambda s: s.nucleus),
    transformation=lambda s: replace(s, is_long=True),
    suffix=RuleSequence(
        [
            Rule[Syllable](
                check_fn=lambda s: s.coda and len(s.coda) == 1 and s.coda[0].val == "x"
            )
        ]
    ),
)

long_position_morphism_4 = Morphism[Syllable](
    target=Rule[Syllable](check_fn=lambda s: s.nucleus),
    transformation=lambda s: replace(s, is_long=True),
    suffix=RuleSequence(
        [
            Rule[Syllable](
                check_fn=lambda s: s.onset
                and len(s.onset) >= 1
                and s.onset[0].val == "x"
            )
        ]
    ),
)
long_position_morphism_5 = Morphism[Syllable](
    target=Rule[Syllable](check_fn=lambda s: s.nucleus and not s.coda),
    transformation=lambda s: replace(s, is_long=True),
    suffix=RuleSequence(
        [Rule[Syllable](check_fn=lambda s: s.onset and len(s.onset) >= 2)]
    ),
)

# Morphism Store
latin_syllable_morphisms = MorphismStore[Syllable](
    [
        long_nature_morphism,
        long_position_morphism_1,
        long_position_morphism_2,
        long_position_morphism_3,
        long_position_morphism_4,
        long_position_morphism_5,
    ]
)

latin_phoneme_morphisms = MorphismStore([])



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="loquax",
    version="0.1.2",
    author="Matthieu Court",
    author_email="matthieu.court@protonmail.com",
    description="A Classical Phonology framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mattlianje/loquax",
    packages=find_packages(exclude=['tests', 'tests.*']),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires='>=3.10',
)


from typing import List, Set

from loquax.abstractions import Language, Phoneme


def get_phonemes(word: str, lang: Language) -> List[Phoneme]:
    def _find_match(word: str, max_len: int, letter_groups: Set[str]) -> str:
        return (
            word[:max_len]
            if max_len > 0 and word[:max_len] in letter_groups
            else _find_match(word, max_len - 1, letter_groups)
            if max_len > 0
            else word[0]
        )

    def _split_token(word: str, symbol_groups: Set[str]) -> List[Phoneme]:
        longest = (
            len(max(lang.constants.symbol_groups, key=len))
            if lang.constants.symbol_groups
            else 0
        )
        return (
            []
            if word == ""
            else [
                Phoneme(_find_match(word, longest, symbol_groups), lang),
                *_split_token(
                    word[len(_find_match(word, longest, symbol_groups)) :],
                    symbol_groups,
                ),
            ]
        )

    return _split_token(word, lang.constants.symbol_groups)



import re
from functools import reduce
from typing import List

from loquax.abstractions import Tokenizer


class LatinTokenizer(Tokenizer):
    """
    Tokenizer for Classical Latin text.
    """

    def tokenize(self, text: str) -> List[str]:
        compose = lambda *functions: reduce(
            lambda f, g: lambda x: f(g(x)), functions, lambda x: x
        )
        transformations = compose(
            str.lower,
            lambda s: re.sub(r"[^\w\s]", "", s),
            lambda s: re.sub(r"\d", "", s),
        )
        return transformations(text).split()



from dataclasses import dataclass
from functools import reduce
from typing import List, Callable

from loquax.abstractions import Language, Syllable
from loquax.abstractions.syllabification import get_syllables_from_token


@dataclass
class Token:
    """
    Class to represent a token or sequence of symbols meaningful to a sentence
    """

    value: str
    language: Language

    @property
    def syllables(self) -> List[Syllable]:
        return get_syllables_from_token(self.value, self.language)

    def to_str(self, ipa: bool = False, scansion: bool = False) -> str:
        syllable_reprs: List[str] = list(
            map(lambda syl: syl.to_str(ipa), self.syllables)
        )
        syllable_repr_str: str = reduce(
            lambda acc, repr: acc + "." + repr, syllable_reprs
        )

        if scansion:
            center_scansion: Callable[
                [str, Syllable], str
            ] = lambda syllable_repr, syllable: syllable.scansion_str(ipa).center(
                len(syllable_repr)
            )

            scansion_reprs: List[str] = list(
                map(center_scansion, syllable_reprs, self.syllables)
            )
            scansion_repr_str: str = "\n" + reduce(
                lambda acc, repr: acc + " " + repr, scansion_reprs
            )

            return syllable_repr_str + scansion_repr_str

        return syllable_repr_str


@dataclass
class Document:
    """
    Class to represent a text or sequence of tokens.
    """

    val: str
    language: Language

    @property
    def tokens(self) -> List[Token]:
        return [
            Token(token, self.language)
            for token in self.language.tokenizer.tokenize(self.val)
        ]

    def to_str(self, ipa: bool = False, scansion: bool = False) -> str:
        join_syllables: Callable[[Token], str] = lambda token: ".".join(
            map(lambda syl: syl.to_str(ipa), token.syllables)
        )

        token_syllable_reprs: List[str] = list(map(join_syllables, self.tokens))
        token_syllable_repr_str: str = reduce(
            lambda acc, repr: acc + "    " + repr, token_syllable_reprs
        )

        if scansion:
            join_scansion: Callable[[Token], str] = lambda token: " ".join(
                map(
                    lambda syl: syl.scansion_str(ipa).center(len(syl.to_str(ipa))),
                    token.syllables,
                )
            )

            token_scansion_reprs: List[str] = list(map(join_scansion, self.tokens))
            token_scansion_repr_str: str = "\n" + reduce(
                lambda acc, repr: acc + "    " + repr, token_scansion_reprs
            )

            return token_syllable_repr_str + token_scansion_repr_str

        return token_syllable_repr_str



import unicodedata

from loquax.abstractions import Phoneme


def has_macron(phoneme: Phoneme) -> bool:
    """Return True if any character in the phoneme's value has a macron accent."""
    return any("macron" in unicodedata.name(c, "").lower() for c in phoneme.val)


def has_trema(phoneme: Phoneme) -> bool:
    """Return True if any character in the phoneme's value has a trema (diaeresis) accent."""
    return any("diaeresis" in unicodedata.name(c, "").lower() for c in phoneme.val)


def has_acute(phoneme: Phoneme) -> bool:
    """Return True if any character in the phoneme's value has an acute accent."""
    return any("acute" in unicodedata.name(c, "").lower() for c in phoneme.val)


def has_grave(phoneme: Phoneme) -> bool:
    """Return True if any character in the phoneme's value has a grave accent."""
    return any("grave" in unicodedata.name(c, "").lower() for c in phoneme.val)


def has_tilde(phoneme: Phoneme) -> bool:
    """Return True if any character in the phoneme's value has a tilde accent."""
    return any("tilde" in unicodedata.name(c, "").lower() for c in phoneme.val)
